@@ -41,6 +41,7 @@ def test_add(rule1, rule2, rs):
     "exp_covs,"
     "exp_preds,"
     "exp_crits,"
+    "exp_zscores,"
     "exp_pred,"
     "exp_crit,"
     "ruleset_pred_crit_fails",
@@ -61,6 +62,8 @@ def test_add(rule1, rule2, rs):
             pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([0.25, 1], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([abs(2**(1/2)*(1.5-1.75)/(0.6875**(1/2))), abs(2**(1/2)*(2-1.75)/(0.6875**(1/2)))],
+                      index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([np.nan, 2, 1.5, 1.75]),
             1.8125 / 3.0,
             False,
@@ -81,6 +84,8 @@ def test_add(rule1, rule2, rs):
             pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([0.25, 1], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([abs(2**(1/2)*(1.5-1.75)/(0.6875**(1/2))), abs(2**(1/2)*(2-1.75)/(0.6875**(1/2)))],
+                      index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([np.nan, np.nan, 1.5, 1.5]),
             0.25,
             False,
@@ -101,6 +106,9 @@ def test_add(rule1, rule2, rs):
             pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([0.25, 1], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([abs(2 ** (1 / 2) * (1.5 - 1.75) / (0.6875 ** (1 / 2))),
+                       abs(2 ** (1 / 2) * (2 - 1.75) / (0.6875 ** (1 / 2)))],
+                      index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([np.nan, np.nan, 1.5, 1.5]),
             0.25,
             False,
@@ -121,6 +129,9 @@ def test_add(rule1, rule2, rs):
             pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([0.25, 1], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([abs(2 ** (1 / 2) * (1.5 - 1.75) / (0.6875 ** (1 / 2))),
+                       abs(2 ** (1 / 2) * (2 - 1.75) / (0.6875 ** (1 / 2)))],
+                      index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([np.nan, 2, 1.5, 1.785714286]),
             0.6224489795918368,
             False,
@@ -141,10 +152,201 @@ def test_add(rule1, rule2, rs):
             pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([0.25, 1], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([abs(2 ** (1 / 2) * (1.5 - 1.75) / (0.6875 ** (1 / 2))),
+                       abs(2 ** (1 / 2) * (2 - 1.75) / (0.6875 ** (1 / 2)))],
+                      index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([np.nan, 2, 1.5, 1.9]),
             2.06 / 3,
             False,
         ),
+        (
+            np.array([[5, 3], [3, 4], [2, np.nan], [2, 4]]),
+            np.array([1, 3, 2, 1]),
+            np.array([[1, 2], [np.nan, 3]]),
+            np.array([1, 2]),
+            [
+                RegressionRule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
+                RegressionRule(HyperrectangleCondition([1], bmins=[4], bmaxs=[5])),
+            ],
+            None,
+            0.75,
+            np.array([0, 1, 1, 1]),
+            pd.DataFrame([[0, 0], [0, 1], [1, 0], [1, 1]], dtype=np.uint8, columns=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([0.25, np.nan], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([abs(2 ** (1 / 2) * (1.5 - 1.75) / (0.6875 ** (1 / 2))),
+                       abs(2 ** (1 / 2) * (2 - 1.75) / (0.6875 ** (1 / 2)))],
+                      index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([np.nan, 2, 1.5, 1.75]),
+            0.25,
+            False,
+        ),
+        (
+            np.array([[5, 3], [3, 4], [2, np.nan], [2, 4]]),
+            np.array([1, 3, 2, 1]),
+            np.array([[1, 2], [np.nan, 3]]),
+            np.array([1, 2]),
+            [
+                RegressionRule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
+                RegressionRule(HyperrectangleCondition([1], bmins=[4], bmaxs=[5])),
+            ],
+            pd.Series([1, 0], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            0.75,
+            np.array([0, 1, 1, 1]),
+            pd.DataFrame([[0, 0], [0, 1], [1, 0], [1, 1]], dtype=np.uint8, columns=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([0.25, np.nan], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([abs(2 ** (1 / 2) * (1.5 - 1.75) / (0.6875 ** (1 / 2))),
+                       abs(2 ** (1 / 2) * (2 - 1.75) / (0.6875 ** (1 / 2)))],
+                      index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([np.nan, np.nan, 1.5, 1.5]),
+            0.25,
+            False,
+        ),
+        (
+            np.array([[5, 3], [3, 4], [2, np.nan], [2, 4]]),
+            np.array([1, 3, 2, 1]),
+            np.array([[1, 2], [np.nan, 3]]),
+            np.array([1, 2]),
+            [
+                RegressionRule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
+                RegressionRule(HyperrectangleCondition([1], bmins=[4], bmaxs=[5])),
+            ],
+            pd.Series([1], index=["X_0 in [1, 2]"]),
+            0.75,
+            np.array([0, 1, 1, 1]),
+            pd.DataFrame([[0, 0], [0, 1], [1, 0], [1, 1]], dtype=np.uint8, columns=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([0.25, np.nan], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([abs(2 ** (1 / 2) * (1.5 - 1.75) / (0.6875 ** (1 / 2))),
+                       abs(2 ** (1 / 2) * (2 - 1.75) / (0.6875 ** (1 / 2)))],
+                      index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([np.nan, np.nan, 1.5, 1.5]),
+            0.25,
+            False,
+        ),
+        (
+            np.array([[5, 3], [3, 4], [2, np.nan], [2, 4]]),
+            np.array([1, 3, 2, 1]),
+            np.array([[1, 2], [np.nan, 3]]),
+            np.array([1, 2]),
+            [
+                RegressionRule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
+                RegressionRule(HyperrectangleCondition([1], bmins=[4], bmaxs=[5])),
+            ],
+            pd.Series([0.75, 1], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            0.75,
+            np.array([0, 1, 1, 1]),
+            pd.DataFrame([[0, 0], [0, 1], [1, 0], [1, 1]], dtype=np.uint8, columns=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([0.25, np.nan], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([abs(2 ** (1 / 2) * (1.5 - 1.75) / (0.6875 ** (1 / 2))),
+                       abs(2 ** (1 / 2) * (2 - 1.75) / (0.6875 ** (1 / 2)))],
+                      index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([np.nan, 2, 1.5, 1.785714286]),
+            0.25,
+            False,
+        ),
+        (
+            np.array([[5, 3], [3, 4], [2, np.nan], [2, 4]]),
+            np.array([1, 3, 2, 1]),
+            np.array([[1, 2], [np.nan, 3]]),
+            np.array([1, 2]),
+            [
+                RegressionRule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
+                RegressionRule(HyperrectangleCondition([1], bmins=[4], bmaxs=[5])),
+            ],
+            "criterion",
+            0.75,
+            np.array([0, 1, 1, 1]),
+            pd.DataFrame([[0, 0], [0, 1], [1, 0], [1, 1]], dtype=np.uint8, columns=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([0.25, np.nan], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([abs(2 ** (1 / 2) * (1.5 - 1.75) / (0.6875 ** (1 / 2))),
+                       abs(2 ** (1 / 2) * (2 - 1.75) / (0.6875 ** (1 / 2)))],
+                      index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([np.nan, np.nan, 1.5, 1.5]),
+            0.25,
+            False,
+        ),
+    ],
+)
+def test_unstacked_fit_regression(
+    clean,
+    xs,
+    y,
+    xs_test,
+    y_test,
+    rule_list,
+    weights,
+    exp_coverage,
+    exp_act,
+    exp_stacked_act,
+    exp_covs,
+    exp_preds,
+    exp_crits,
+    exp_zscores,
+    exp_pred,
+    exp_crit,
+    ruleset_pred_crit_fails,
+):
+    res = RuleSet(rule_list, compute_activation=True, stack_activation=True)
+    res.fit(y, xs)
+    if ruleset_pred_crit_fails:
+        with pytest.raises(ValueError) as e:
+            res.eval(
+                y_test,
+                xs_test,
+                weights=weights,
+                criterion_method="mse" if issubclass(res._rule_type, RegressionRule) else "success_rate",
+            )
+            assert "No rules had non-zero/non-NaN weights" in str(e)
+        return
+    else:
+        res.eval(
+            y_test,
+            xs_test,
+            weights=weights,
+            criterion_method="mse" if issubclass(res._rule_type, RegressionRule) else "success_rate",
+        )
+    assert res.ruleset_coverage == exp_coverage
+    np.testing.assert_equal(res.activation, exp_act)
+    pd.testing.assert_frame_equal(res.stacked_activations, exp_stacked_act)
+    for r in res:
+        assert r.coverage == exp_covs[str(r.condition)]
+        assert r.prediction == exp_preds[str(r.condition)]
+        assert round(r.zscore, 6) == round(exp_zscores[str(r.condition)], 6)
+        if np.isnan(r.criterion):
+            assert np.isnan(exp_crits[str(r.condition)])
+        else:
+            assert r.criterion == exp_crits[str(r.condition)]
+    prediction = res.predict(xs=xs, weights=weights)
+    pd.testing.assert_series_equal(prediction, exp_pred)
+    assert round(res.criterion, 6) == round(exp_crit, 6)
+
+
+@pytest.mark.parametrize(
+    "xs,"
+    "y,"
+    "xs_test,"
+    "y_test,"
+    "rule_list,"
+    "weights,"
+    "exp_coverage,"
+    "exp_act,"
+    "exp_stacked_act,"
+    "exp_covs,"
+    "exp_preds,"
+    "exp_crits,"
+    "exp_pred,"
+    "exp_crit,"
+    "ruleset_pred_crit_fails",
+    [
         (
             np.array([[5, 3], [5, 4], [2, np.nan], [2, 4], [2, 6]]),
             np.array([1, 3, 2, 2, 3]),
@@ -270,106 +472,6 @@ def test_add(rule1, rule2, rs):
             1.0,
             False,
         ),
-        (
-            np.array([[5, 3], [3, 4], [2, np.nan], [2, 4]]),
-            np.array([1, 3, 2, 1]),
-            np.array([[1, 2], [np.nan, 3]]),
-            np.array([1, 2]),
-            [
-                RegressionRule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
-                RegressionRule(HyperrectangleCondition([1], bmins=[4], bmaxs=[5])),
-            ],
-            None,
-            0.75,
-            np.array([0, 1, 1, 1]),
-            pd.DataFrame([[0, 0], [0, 1], [1, 0], [1, 1]], dtype=np.uint8, columns=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([0.25, np.nan], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([np.nan, 2, 1.5, 1.75]),
-            0.25,
-            False,
-        ),
-        (
-            np.array([[5, 3], [3, 4], [2, np.nan], [2, 4]]),
-            np.array([1, 3, 2, 1]),
-            np.array([[1, 2], [np.nan, 3]]),
-            np.array([1, 2]),
-            [
-                RegressionRule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
-                RegressionRule(HyperrectangleCondition([1], bmins=[4], bmaxs=[5])),
-            ],
-            pd.Series([1, 0], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            0.75,
-            np.array([0, 1, 1, 1]),
-            pd.DataFrame([[0, 0], [0, 1], [1, 0], [1, 1]], dtype=np.uint8, columns=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([0.25, np.nan], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([np.nan, np.nan, 1.5, 1.5]),
-            0.25,
-            False,
-        ),
-        (
-            np.array([[5, 3], [3, 4], [2, np.nan], [2, 4]]),
-            np.array([1, 3, 2, 1]),
-            np.array([[1, 2], [np.nan, 3]]),
-            np.array([1, 2]),
-            [
-                RegressionRule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
-                RegressionRule(HyperrectangleCondition([1], bmins=[4], bmaxs=[5])),
-            ],
-            pd.Series([1], index=["X_0 in [1, 2]"]),
-            0.75,
-            np.array([0, 1, 1, 1]),
-            pd.DataFrame([[0, 0], [0, 1], [1, 0], [1, 1]], dtype=np.uint8, columns=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([0.25, np.nan], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([np.nan, np.nan, 1.5, 1.5]),
-            0.25,
-            False,
-        ),
-        (
-            np.array([[5, 3], [3, 4], [2, np.nan], [2, 4]]),
-            np.array([1, 3, 2, 1]),
-            np.array([[1, 2], [np.nan, 3]]),
-            np.array([1, 2]),
-            [
-                RegressionRule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
-                RegressionRule(HyperrectangleCondition([1], bmins=[4], bmaxs=[5])),
-            ],
-            pd.Series([0.75, 1], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            0.75,
-            np.array([0, 1, 1, 1]),
-            pd.DataFrame([[0, 0], [0, 1], [1, 0], [1, 1]], dtype=np.uint8, columns=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([0.25, np.nan], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([np.nan, 2, 1.5, 1.785714286]),
-            0.25,
-            False,
-        ),
-        (
-            np.array([[5, 3], [3, 4], [2, np.nan], [2, 4]]),
-            np.array([1, 3, 2, 1]),
-            np.array([[1, 2], [np.nan, 3]]),
-            np.array([1, 2]),
-            [
-                RegressionRule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
-                RegressionRule(HyperrectangleCondition([1], bmins=[4], bmaxs=[5])),
-            ],
-            "criterion",
-            0.75,
-            np.array([0, 1, 1, 1]),
-            pd.DataFrame([[0, 0], [0, 1], [1, 0], [1, 1]], dtype=np.uint8, columns=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([0.25, np.nan], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([np.nan, np.nan, 1.5, 1.5]),
-            0.25,
-            False,
-        ),
         (
             np.array([[5, 3], [5, 4], [2, np.nan], [2, 4], [2, 6]]),
             np.array([1, 3, 2, 2, 3]),
@@ -501,7 +603,7 @@ def test_add(rule1, rule2, rs):
         ),
     ],
 )
-def test_unstacked_fit(
+def test_unstacked_fit_classification(
     clean,
     xs,
     y,
@@ -566,6 +668,7 @@ def test_unstacked_fit(
     "exp_covs,"
     "exp_preds,"
     "exp_crits,"
+    "exp_zscores,"
     "exp_pred,"
     "exp_crit,"
     "ruleset_pred_crit_fails",
@@ -586,6 +689,8 @@ def test_unstacked_fit(
             pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([0.25, 1], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([abs(2**(1/2)*(1.5-1.75)/(0.6875**(1/2))), abs(2**(1/2)*(2-1.75)/(0.6875**(1/2)))],
+                      index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([np.nan, 2, 1.5, 1.75]),
             1.8125 / 3.0,
             False,
@@ -606,6 +711,8 @@ def test_unstacked_fit(
             pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([0.25, 1], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([abs(2**(1/2)*(1.5-1.75)/(0.6875**(1/2))), abs(2**(1/2)*(2-1.75)/(0.6875**(1/2)))],
+                      index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([np.nan, np.nan, 1.5, 1.5]),
             0.25,
             False,
@@ -626,6 +733,8 @@ def test_unstacked_fit(
             pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([0.25, 1], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([abs(2**(1/2)*(1.5-1.75)/(0.6875**(1/2))), abs(2**(1/2)*(2-1.75)/(0.6875**(1/2)))],
+                      index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([np.nan, np.nan, 1.5, 1.5]),
             0.25,
             False,
@@ -646,6 +755,8 @@ def test_unstacked_fit(
             pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([0.25, 1], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([abs(2**(1/2)*(1.5-1.75)/(0.6875**(1/2))), abs(2**(1/2)*(2-1.75)/(0.6875**(1/2)))],
+                      index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([np.nan, 2, 1.5, 1.785714286]),
             0.6224489795918368,
             False,
@@ -666,10 +777,197 @@ def test_unstacked_fit(
             pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([0.25, 1], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([abs(2**(1/2)*(1.5-1.75)/(0.6875**(1/2))), abs(2**(1/2)*(2-1.75)/(0.6875**(1/2)))],
+                      index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
             pd.Series([np.nan, 2, 1.5, 1.9]),
             2.06 / 3,
             False,
         ),
+        (
+            np.array([[5, 3], [3, 4], [2, np.nan], [2, 4]]),
+            np.array([1, 3, 2, 1]),
+            np.array([[1, 2], [np.nan, 3]]),
+            np.array([1, 2]),
+            [
+                RegressionRule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
+                RegressionRule(HyperrectangleCondition([1], bmins=[4], bmaxs=[5])),
+            ],
+            None,
+            0.75,
+            np.array([0, 1, 1, 1]),
+            pd.DataFrame([[0, 0], [0, 1], [1, 0], [1, 1]], dtype=np.uint8, columns=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([0.25, np.nan], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([abs(2**(1/2)*(1.5-1.75)/(0.6875**(1/2))), abs(2**(1/2)*(2-1.75)/(0.6875**(1/2)))],
+                      index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([np.nan, 2, 1.5, 1.75]),
+            0.25,
+            False,
+        ),
+        (
+            np.array([[5, 3], [3, 4], [2, np.nan], [2, 4]]),
+            np.array([1, 3, 2, 1]),
+            np.array([[1, 2], [np.nan, 3]]),
+            np.array([1, 2]),
+            [
+                RegressionRule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
+                RegressionRule(HyperrectangleCondition([1], bmins=[4], bmaxs=[5])),
+            ],
+            pd.Series([1, 0], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            0.75,
+            np.array([0, 1, 1, 1]),
+            pd.DataFrame([[0, 0], [0, 1], [1, 0], [1, 1]], dtype=np.uint8, columns=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([0.25, np.nan], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([abs(2**(1/2)*(1.5-1.75)/(0.6875**(1/2))), abs(2**(1/2)*(2-1.75)/(0.6875**(1/2)))],
+                      index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([np.nan, np.nan, 1.5, 1.5]),
+            0.25,
+            False,
+        ),
+        (
+            np.array([[5, 3], [3, 4], [2, np.nan], [2, 4]]),
+            np.array([1, 3, 2, 1]),
+            np.array([[1, 2], [np.nan, 3]]),
+            np.array([1, 2]),
+            [
+                RegressionRule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
+                RegressionRule(HyperrectangleCondition([1], bmins=[4], bmaxs=[5])),
+            ],
+            pd.Series([1], index=["X_0 in [1, 2]"]),
+            0.75,
+            np.array([0, 1, 1, 1]),
+            pd.DataFrame([[0, 0], [0, 1], [1, 0], [1, 1]], dtype=np.uint8, columns=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([0.25, np.nan], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([abs(2**(1/2)*(1.5-1.75)/(0.6875**(1/2))), abs(2**(1/2)*(2-1.75)/(0.6875**(1/2)))],
+                      index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([np.nan, np.nan, 1.5, 1.5]),
+            0.25,
+            False,
+        ),
+        (
+            np.array([[5, 3], [3, 4], [2, np.nan], [2, 4]]),
+            np.array([1, 3, 2, 1]),
+            np.array([[1, 2], [np.nan, 3]]),
+            np.array([1, 2]),
+            [
+                RegressionRule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
+                RegressionRule(HyperrectangleCondition([1], bmins=[4], bmaxs=[5])),
+            ],
+            pd.Series([0.75, 1], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            0.75,
+            np.array([0, 1, 1, 1]),
+            pd.DataFrame([[0, 0], [0, 1], [1, 0], [1, 1]], dtype=np.uint8, columns=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([0.25, np.nan], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([abs(2**(1/2)*(1.5-1.75)/(0.6875**(1/2))), abs(2**(1/2)*(2-1.75)/(0.6875**(1/2)))],
+                      index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([np.nan, 2, 1.5, 1.785714286]),
+            0.25,
+            False,
+        ),
+        (
+            np.array([[5, 3], [3, 4], [2, np.nan], [2, 4]]),
+            np.array([1, 3, 2, 1]),
+            np.array([[1, 2], [np.nan, 3]]),
+            np.array([1, 2]),
+            [
+                RegressionRule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
+                RegressionRule(HyperrectangleCondition([1], bmins=[4], bmaxs=[5])),
+            ],
+            "criterion",
+            0.75,
+            np.array([0, 1, 1, 1]),
+            pd.DataFrame([[0, 0], [0, 1], [1, 0], [1, 1]], dtype=np.uint8, columns=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([0.25, np.nan], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([abs(2**(1/2)*(1.5-1.75)/(0.6875**(1/2))), abs(2**(1/2)*(2-1.75)/(0.6875**(1/2)))],
+                      index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
+            pd.Series([np.nan, np.nan, 1.5, 1.5]),
+            0.25,
+            False,
+        ),
+    ],
+)
+def test_stacked_fit_regression(
+    clean_for_stacked_fit,
+    xs,
+    y,
+    xs_test,
+    y_test,
+    rule_list,
+    weights,
+    exp_coverage,
+    exp_act,
+    exp_stacked_act,
+    exp_covs,
+    exp_preds,
+    exp_crits,
+    exp_zscores,
+    exp_pred,
+    exp_crit,
+    ruleset_pred_crit_fails,
+):
+    res = RuleSet(rule_list, compute_activation=True, stack_activation=True)
+    res.fit(y, xs)
+    assert res.train_set_size == len(y)
+    if ruleset_pred_crit_fails:
+        with pytest.raises(ValueError) as e:
+            res.eval(
+                y_test,
+                xs_test,
+                weights=weights,
+                criterion_method="mse" if issubclass(res._rule_type, RegressionRule) else "success_rate",
+            )
+            assert "No rules had non-zero/non-NaN weights" in str(e)
+        return
+    else:
+        res.eval(
+            y_test,
+            xs_test,
+            weights=weights,
+            criterion_method="mse" if issubclass(res._rule_type, RegressionRule) else "success_rate",
+        )
+        assert res.test_set_size == len(y_test)
+    assert res.ruleset_coverage == exp_coverage
+    np.testing.assert_equal(res.activation, exp_act)
+    pd.testing.assert_frame_equal(res.stacked_activations, exp_stacked_act)
+    for r in res:
+        assert r.coverage == exp_covs[str(r.condition)]
+        assert r.prediction == exp_preds[str(r.condition)]
+        assert round(r.zscore, 6) == round(exp_zscores[str(r.condition)], 6)
+        if np.isnan(r.criterion):
+            assert np.isnan(exp_crits[str(r.condition)])
+        else:
+            assert r.criterion == exp_crits[str(r.condition)]
+    prediction = res.predict(xs=xs, weights=weights)
+    pd.testing.assert_series_equal(prediction, exp_pred)
+    assert round(res.criterion, 6) == round(exp_crit, 6)
+
+
+@pytest.mark.parametrize(
+    "xs,"
+    "y,"
+    "xs_test,"
+    "y_test,"
+    "rule_list,"
+    "weights,"
+    "exp_coverage,"
+    "exp_act,"
+    "exp_stacked_act,"
+    "exp_covs,"
+    "exp_preds,"
+    "exp_crits,"
+    "exp_pred,"
+    "exp_crit,"
+    "ruleset_pred_crit_fails",
+    [
         (
             np.array([[5, 3], [5, 4], [2, np.nan], [2, 4], [2, 6]]),
             np.array([1, 3, 2, 2, 3]),
@@ -794,107 +1092,7 @@ def test_unstacked_fit(
             pd.Series([np.nan, "c", "b", "b", np.nan]),
             1.0,
             False,
-        ),
-        (
-            np.array([[5, 3], [3, 4], [2, np.nan], [2, 4]]),
-            np.array([1, 3, 2, 1]),
-            np.array([[1, 2], [np.nan, 3]]),
-            np.array([1, 2]),
-            [
-                RegressionRule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
-                RegressionRule(HyperrectangleCondition([1], bmins=[4], bmaxs=[5])),
-            ],
-            None,
-            0.75,
-            np.array([0, 1, 1, 1]),
-            pd.DataFrame([[0, 0], [0, 1], [1, 0], [1, 1]], dtype=np.uint8, columns=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([0.25, np.nan], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([np.nan, 2, 1.5, 1.75]),
-            0.25,
-            False,
-        ),
-        (
-            np.array([[5, 3], [3, 4], [2, np.nan], [2, 4]]),
-            np.array([1, 3, 2, 1]),
-            np.array([[1, 2], [np.nan, 3]]),
-            np.array([1, 2]),
-            [
-                RegressionRule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
-                RegressionRule(HyperrectangleCondition([1], bmins=[4], bmaxs=[5])),
-            ],
-            pd.Series([1, 0], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            0.75,
-            np.array([0, 1, 1, 1]),
-            pd.DataFrame([[0, 0], [0, 1], [1, 0], [1, 1]], dtype=np.uint8, columns=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([0.25, np.nan], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([np.nan, np.nan, 1.5, 1.5]),
-            0.25,
-            False,
-        ),
-        (
-            np.array([[5, 3], [3, 4], [2, np.nan], [2, 4]]),
-            np.array([1, 3, 2, 1]),
-            np.array([[1, 2], [np.nan, 3]]),
-            np.array([1, 2]),
-            [
-                RegressionRule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
-                RegressionRule(HyperrectangleCondition([1], bmins=[4], bmaxs=[5])),
-            ],
-            pd.Series([1], index=["X_0 in [1, 2]"]),
-            0.75,
-            np.array([0, 1, 1, 1]),
-            pd.DataFrame([[0, 0], [0, 1], [1, 0], [1, 1]], dtype=np.uint8, columns=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([0.25, np.nan], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([np.nan, np.nan, 1.5, 1.5]),
-            0.25,
-            False,
-        ),
-        (
-            np.array([[5, 3], [3, 4], [2, np.nan], [2, 4]]),
-            np.array([1, 3, 2, 1]),
-            np.array([[1, 2], [np.nan, 3]]),
-            np.array([1, 2]),
-            [
-                RegressionRule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
-                RegressionRule(HyperrectangleCondition([1], bmins=[4], bmaxs=[5])),
-            ],
-            pd.Series([0.75, 1], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            0.75,
-            np.array([0, 1, 1, 1]),
-            pd.DataFrame([[0, 0], [0, 1], [1, 0], [1, 1]], dtype=np.uint8, columns=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([0.25, np.nan], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([np.nan, 2, 1.5, 1.785714286]),
-            0.25,
-            False,
-        ),
-        (
-            np.array([[5, 3], [3, 4], [2, np.nan], [2, 4]]),
-            np.array([1, 3, 2, 1]),
-            np.array([[1, 2], [np.nan, 3]]),
-            np.array([1, 2]),
-            [
-                RegressionRule(HyperrectangleCondition([0], bmins=[1], bmaxs=[2])),
-                RegressionRule(HyperrectangleCondition([1], bmins=[4], bmaxs=[5])),
-            ],
-            "criterion",
-            0.75,
-            np.array([0, 1, 1, 1]),
-            pd.DataFrame([[0, 0], [0, 1], [1, 0], [1, 1]], dtype=np.uint8, columns=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([0.5, 0.5], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([1.5, 2], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([0.25, np.nan], index=["X_0 in [1, 2]", "X_1 in [4, 5]"]),
-            pd.Series([np.nan, np.nan, 1.5, 1.5]),
-            0.25,
-            False,
-        ),
+            ),
         (
             np.array([[5, 3], [5, 4], [2, np.nan], [2, 4], [2, 6]]),
             np.array([1, 3, 2, 2, 3]),
@@ -1026,7 +1224,7 @@ def test_unstacked_fit(
         ),
     ],
 )
-def test_stacked_fit(
+def test_stacked_fit_classification(
     clean_for_stacked_fit,
     xs,
     y,
