@@ -463,7 +463,7 @@ def test_thresholds_fit(
     if not stacked_fit:
         RuleSet.STACKED_FIT = False
     res.fit(y, xs)
-    res.eval(y, xs, criterion_method="mse" if issubclass(res.rule_type, RegressionRule) else "success_rate")
+    res.eval(y, xs, criterion_method="mse" if issubclass(res._rule_type, RegressionRule) else "success_rate")
     assert res.ruleset_coverage == coverage
     np.testing.assert_equal(res.activation, exp_act)
     pd.testing.assert_frame_equal(res.stacked_activations, exp_stacked_act)
@@ -476,7 +476,7 @@ def test_thresholds_fit(
     RegressionRule.SET_THRESHOLDS("tests/unit/data/thresholds.json")
 
     res.fit(y, xs)
-    res.eval(y, xs, criterion_method="mse" if issubclass(res.rule_type, RegressionRule) else "success_rate")
+    res.eval(y, xs, criterion_method="mse" if issubclass(res._rule_type, RegressionRule) else "success_rate")
     pd.testing.assert_frame_equal(res.stacked_activations, exp_stacked_act_2)
     for r in res:
         assert r.coverage == exp_covs_2[str(r.condition)]
